/*
 * Copyright 2011 - 2017 Intel Corporation
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if defined(HAVE_CONFIG_H)
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <cstddef>
#include <log/log.h>

#include "mdadm_config.h"
#include "filesystem.h"
#include "utils.h"

namespace {
    const char mdadmMonitorPid[] = "/var/run/mdadm/autorebuild.pid";

    static String stdConfig = "# This config has been automatically generated by SSI API library\n"
                              "# for use with CIM Provider. Please do not edit\n"
                              "\n"
                              "MAILADDR root\n"
                              "AUTO +imsm +1.x -all\n"
                              "policy domain=global metadata=imsm path=* action=spare-same-slot\n";

    /* */
    bool correct_config(String &config) {
        return config + "\n" == stdConfig;
    }

    /* */
    int write_config(String file, String &config)
    {
        File attr = file;
        try {
            attr << config;
        } catch (...) {
            dlog("Warning: file " + file + " could not be written");
            return 1;
        }
        return 0;
    }

    /* */
    void backup_config(String &config)
    {
        write_config("/etc/mdadm.conf.original", config);
    }

    /* */
    bool monitor_running()
    {
        File attr = String(mdadmMonitorPid);
        String pid;
        try {
            attr >> pid;
        } catch (...) {
            return false;
        }

        String buffer;
        String command = "ps " + pid;
        if (shell_cap(command, buffer) == 0) {
            try {
                buffer.find("mdadm");
            } catch (...) {
                return false;
            }
            try {
                buffer.find("ssimsg");
            } catch (...) {
                return false;
            }

            return true;
        }

        return false;
    }

    /* */
    bool start_monitor()
    {
        if (monitor_running()) {
            return true;
        }

        String s = "ssimsg";

        try {
            if (shell_cap("whereis -b ssimsg", s) == 0) {
                s = s.reverse_after(" ");
            }
        } catch (...) {
        }

        dlog("starting Monitor...");
        return shell_command("mdadm --monitor -y --scan --daemonise -p " + s) == 0;
    }

    /* */
    void stop_monitor()
    {
        File attr = String(mdadmMonitorPid);
        String pid;
        try {
            attr >> pid;
            if (shell_command("kill -n 15 " + pid) == 0) {
                dlog("killed Monitor: " + pid);
            }
        } catch (...) {
            /* No monitor to kill */
        }
    }

    /* */
    bool restart_monitor()
    {
        stop_monitor();

        return start_monitor();
    }
}

/* */
void configuration_init()
{
    File attr = String(MDADM_CONFIG_PATH);
    String config;
    bool configOk = false;
    bool backup = true;

    /*check mdadm.conf*/
    try {
        attr >> config;
        configOk = correct_config(config);
    } catch (Exception ex) {
        if (ex != E_NOT_FOUND)
            dlog("Warning: mdadm config file cannot be read, new one will be written");
            backup = false;
    }

    if (configOk && monitor_running()) {
        dlog("Configuration correct");
        return;
    }

    if (!configOk) {
        if (backup)
            backup_config(config);
        if (write_config(MDADM_CONFIG_PATH, stdConfig) != 0)
            dlog("Warning: failed to update mdadm.conf");
    }

    if (restart_monitor()) {
        dlog("Monitor restarted successfully");
    } else {
        dlog("Error starting Monitor");
    }
}

/* */
void check_configuration()
{
    configuration_init();
}

/* */
void configuration_fini()
{
    stop_monitor();
}

void get_mdadm_version(char *buffer, size_t size)
{
    String ver;
    if (!buffer || size == 0) {
        return;
    }

    if (shell_cap("mdadm --version 2>&1", ver) == 0) {
        ver = ver.between("- v", " ");
        check_dots(ver);
        ver.get(buffer, size);
    }
}
/* ex: set tabstop=4 softtabstop=4 shiftwidth=4 textwidth=98 expandtab: */
